/* db_form.c */

#include <Pilot.h>

#include "callback.h"

#include "pz.h"
#include "pz_rcp.h"


/*
 * macros
 */

#define MAX_LIST_HEIGHT 14
#define LIST_HEIGHT(x) ((x) < MAX_LIST_HEIGHT ? (x) : MAX_LIST_HEIGHT)


/*
 * static variables
 */

#ifndef UNPZ
static struct db_spec *arch_db_list;
static int num_arch_dbs;

static char arch_db_name[dmDBNameLength];
static Boolean arch_copy_protect;
#endif


/*
 * database list
 */

static void update_db_list(Word update_code)
{
    int sel, i;

    DrawBitmapOver(WATCH_ICON_ID, DB_TABLE_ID);

    if (update_code != UPDATE_DB_SORT
#ifndef UNPZ
     || prefs.sort_by == sort_by_none
#endif
     || db_list == NULL) {
	free_db_list();
#ifndef UNPZ
	gen_db_list(prefs.where, prefs.kind);
#else
	gen_db_list(where_all, kind_archive);
#endif
    }
#ifndef UNPZ
    if (prefs.sort_by != sort_by_none) {
	sort_db_list(prefs.sort_by);
    }
#else
    sort_db_list(sort_by_name);
#endif

    sel = -1;
    for (i = 0; i < num_dbs; ++i) {
	if (db_list[i].spec.card_no == prefs.last_selection.card_no
	 && db_list[i].spec.db_id == prefs.last_selection.db_id) {
	    sel = i;
	    break;
	}
    }

    if (update_code == UPDATE_DB_LIST) {
	set_table_items(num_dbs, prefs.last_db_top, sel);
    } else {
	set_table_items(num_dbs, 0, sel);
    }
    prefs.last_db_top = get_table_top();
}

static struct db_rec *get_selected_db(void)
{
    int sel;

    sel = get_table_selection();
    if (sel < 0 || sel >= num_dbs) {
	return NULL;
    }

    return &db_list[sel];
}


/*
 * compress dialog
 */

#ifndef UNPZ
static Int compare(struct db_spec *p1, struct db_spec *p2, Long dummy)
{
    char name1[dmDBNameLength], name2[dmDBNameLength];
    Int result;

    CALLBACK_PROLOGUE

    DmDatabaseInfo(p1->card_no, p1->db_id, name1,
	    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
    DmDatabaseInfo(p2->card_no, p2->db_id, name2,
	    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);

    result = StrCaselessCompare(name1, name2);

    CALLBACK_EPILOGUE

    return result;
}

static void gen_arch_db_list(void)
{
    Boolean new_search;
    DmSearchStateType state;
    int cnt;
    UInt card_no, attr;
    LocalID db_id;
    struct db_rec *p;

    cnt = 0;
    for (new_search = true;
	 DmGetNextDatabaseByTypeCreator(new_search, &state,
		PZ_DB_TYPE, PZ_CREATOR, false, &card_no, &db_id) == noErr;
	 new_search = false) {
	++cnt;
    }

    if (cnt == 0
     || (arch_db_list = MemPtrNew(sizeof(struct db_spec) * cnt)) == NULL) {
	return;
    }

    p = get_selected_db();

    cnt = 0;
    for (new_search = true;
	 DmGetNextDatabaseByTypeCreator(new_search, &state,
		PZ_DB_TYPE, PZ_CREATOR, false, &card_no, &db_id) == noErr;
	 new_search = false) {
	DmDatabaseInfo(card_no, db_id, NULL, &attr,
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
	if ((card_no == p->spec.card_no && db_id == p->spec.db_id)
	 || (attr & dmHdrAttrReadOnly) != 0
	 || (arch_copy_protect && (attr & dmHdrAttrCopyPrevention) == 0)) {
	    continue;
	}
	arch_db_list[cnt].card_no = card_no;
	arch_db_list[cnt].db_id = db_id;
	++cnt;
    }

    num_arch_dbs = cnt;
    SysQSort(arch_db_list, num_arch_dbs,
	    sizeof(struct db_spec), (CmpFuncPtr) compare, NULL);
}

static void free_arch_db_list(void)
{
    if (arch_db_list != NULL) {
	MemPtrFree(arch_db_list);
    }

    arch_db_list = NULL;
    num_arch_dbs = 0;
}

static Boolean get_arch_db_name(Word index, char *name)
{
    if (index < num_arch_dbs) {
	DmDatabaseInfo(
		arch_db_list[index].card_no, arch_db_list[index].db_id, name,
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);

	return true;
    } else {
	return false;
    }
}

static void draw_arch_db_name(Word index, RectanglePtr bounds, CharPtr *dummy)
{
    char name[dmDBNameLength];

    CALLBACK_PROLOGUE

    if (get_arch_db_name(index, name)) {
	draw_fit_chars(name, 0, bounds->extent.x, bounds, false);
    }

    CALLBACK_EPILOGUE
}

static void set_new_or_add(Boolean new)
{
    FormPtr form;

    form = FrmGetActiveForm();
    if (new) {
	FrmSetControlGroupSelection(form, COMPRESS_NEW_OR_ADD, COMPRESS_NEW_ID);
	FrmSetFocus(form, FrmGetObjectIndex(form, COMPRESS_FIELD_ID));
    } else {
	FrmSetControlGroupSelection(form, COMPRESS_NEW_OR_ADD, COMPRESS_ADD_ID);
	FrmSetFocus(form, noFocus);
    }
}

static Boolean handle_compress_form_event(EventPtr event)
{
    Boolean handled = false;

    CALLBACK_PROLOGUE

    switch (event->eType) {
    case fldEnterEvent:
	set_new_or_add(true);
	break;

    case ctlSelectEvent:
	switch (event->data.ctlSelect.controlID) {
	case COMPRESS_NEW_ID:
	case COMPRESS_NEW_CP_ID:
	    set_new_or_add(true);
	    handled = true;
	    break;

	case COMPRESS_ADD_ID:
	    set_new_or_add(false);
	    handled = true;
	    break;
	}
	break;

    case popSelectEvent:
	if (event->data.popSelect.listID == COMPRESS_ADD_LIST_ID
	 && get_arch_db_name(event->data.popSelect.selection, arch_db_name)) {
	    CtlSetLabel(event->data.popSelect.controlP, arch_db_name);
	    set_new_or_add(false);
	    prefs.last_archive = arch_db_list[event->data.popSelect.selection];
	    handled = true;
	}
	break;

    default:
	break;
    }

    CALLBACK_EPILOGUE

    return handled;
}

static Word do_compress_dialog(char *name, Boolean *new,
	Boolean *copy_protect, struct db_spec *existing, Boolean *time)
{
    FormPtr form;
    ControlPtr checkbox;
    FieldPtr field;
    VoidHand h;
    char *p;
    ListPtr add_list;
    Word i, button;

    form = FrmInitForm(COMPRESS_DIALOG_ID);

    arch_copy_protect = *copy_protect;
    checkbox = GetObjectPtrIn(form, COMPRESS_NEW_CP_ID);
    CtlSetValue(checkbox, arch_copy_protect);
    CtlSetEnabled(checkbox, ! arch_copy_protect);

    field = GetObjectPtrIn(form, COMPRESS_FIELD_ID);
    h = MemHandleNew(dmDBNameLength);
    if (h != NULL) {
	p = MemHandleLock(h);
	StrNCopy(p, name, dmDBNameLength - 1);
	p[dmDBNameLength - 1] = '\0';
	MemHandleUnlock(h);
	FldSetTextHandle(field, (Handle) h);
    }
    FrmSetFocus(form, FrmGetObjectIndex(form, COMPRESS_FIELD_ID));

    add_list = GetObjectPtrIn(form, COMPRESS_ADD_LIST_ID);
    LstSetSelection(add_list, noListSelection);
    gen_arch_db_list();
    if (num_arch_dbs > 0) {
	LstSetDrawFunction(add_list, draw_arch_db_name);
	LstSetListChoices(add_list, NULL, num_arch_dbs);
	LstSetHeight(add_list, LIST_HEIGHT(num_arch_dbs));
	for (i = 0; i < num_arch_dbs; ++i) {
	    if (arch_db_list[i].card_no == prefs.last_archive.card_no
	     && arch_db_list[i].db_id == prefs.last_archive.db_id
	     && get_arch_db_name(i, arch_db_name)) {
		CtlSetLabel(
			GetObjectPtrIn(form, COMPRESS_ADD_POPUP_ID),
			arch_db_name);
		LstSetSelection(add_list, i);
		break;
	    }
	}
    }

    FrmSetEventHandler(form, handle_compress_form_event);
    button = FrmDoDialog(form);

    if (button == COMPRESS_OK_ID) {
	*new = CtlGetValue(GetObjectPtrIn(form, COMPRESS_NEW_ID));
	if (*new) {
	    p = FldGetTextPtr(field);
	    if (p != NULL) {
		StrNCopy(name, p, dmDBNameLength - 1);
		name[dmDBNameLength - 1] = '\0';
	    } else {
		button = COMPRESS_CANCEL_ID;
	    }
	} else {
	    i = LstGetSelection(add_list);
	    if (i >= 0 && i < num_arch_dbs) {
		*existing = arch_db_list[i];
	    } else {
		button = COMPRESS_CANCEL_ID;
	    }
	}

	*copy_protect = CtlGetValue(GetObjectPtrIn(form, COMPRESS_NEW_CP_ID));
	*time = CtlGetValue(GetObjectPtrIn(form, COMPRESS_TIME_ID));
    }

    FrmDeleteForm(form);
    free_arch_db_list();

    return button;
}
#endif


/*
 * compressing
 */

#ifndef UNPZ
static int do_compress_new(UInt arch_card_no, const char *name,
	Boolean copy_protect, UInt card_no, LocalID db_id,
	ULong *total_size, ULong *out_size)
{
    LocalID arch_id;
    UInt attr;
    DmOpenRef arch_db;
    int err;

    err = DmCreateDatabase(arch_card_no, name, PZ_CREATOR, PZ_DB_TYPE, false);
    if (err != noErr) {
	return err;	/* DmCreateDatabase() failed */
    }

    arch_id = DmFindDatabase(arch_card_no, (CharPtr) name);
    if (arch_id == NULL) {
	return -1;	/* DmFindDatabase() failed */
    }

    if (copy_protect) {
	DmDatabaseInfo(arch_card_no, arch_id, NULL, &attr,
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
	attr |= dmHdrAttrCopyPrevention;
	DmSetDatabaseInfo(arch_card_no, arch_id, NULL, &attr,
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
    }

    arch_db = DmOpenDatabase(arch_card_no, arch_id, dmModeReadWrite);
    if (arch_db == NULL) {
	return -1;	/* DmOpenDatabase(dmModeReadWrite) failed */
    }

    err = compress_db(arch_db, dmMaxRecordIndex,
	    card_no, db_id, total_size, out_size, show_progress);
    if (err != noErr) {
	DmCloseDatabase(arch_db);
	DmDeleteDatabase(arch_card_no, arch_id);
	return err;	/* compress_db() failed */
    }

    err = DmCloseDatabase(arch_db);
    if (err != noErr) {
	return err;	/* DmCloseDatabase() failed */
    }

    prefs.last_archive.card_no = arch_card_no;
    prefs.last_archive.db_id = arch_id;

    return noErr;
}

static int do_compress_add(UInt arch_card_no, LocalID arch_id,
	Boolean copy_protect, UInt card_no, LocalID db_id,
	ULong *total_size, ULong *out_size)
{
    UInt attr;
    DmOpenRef arch_db;
    int err;

    if (copy_protect) {
	DmDatabaseInfo(arch_card_no, arch_id, NULL, &attr,
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
	if ((attr & dmHdrAttrCopyPrevention) == 0) {
	    return -1;	/* dmHdrAttrCopyPrevention is off... */
	}
    }

    arch_db = DmOpenDatabase(arch_card_no, arch_id, dmModeReadWrite);
    if (arch_db == NULL) {
	return -1;	/* DmOpenDatabase(dmModeReadWrite) failed */
    }

    err = compress_db(arch_db, dmMaxRecordIndex,
	    card_no, db_id, total_size, out_size, show_progress);
    if (err != noErr) {
	DmCloseDatabase(arch_db);
	return err;	/* compress_db() failed */
    }

    err = DmCloseDatabase(arch_db);
    if (err != noErr) {
	return err;	/* DmCloseDatabase() failed */
    }

    prefs.last_archive.card_no = arch_card_no;
    prefs.last_archive.db_id = arch_id;

    return noErr;
}

static void do_compress(void)
{
    struct db_rec *p;
    char db_name[dmDBNameLength], arch_name[dmDBNameLength];
    UInt attr;
    Boolean new, copy_protect, time;
    struct db_spec existing;
    Word button;
    LocalID db_id;
    FormPtr saved_form, form;
    ULong start_ticks, end_ticks, total_size, out_size;
    char buf1[12], buf2[12], buf3[12];
    int err;

    p = get_selected_db();
    if (p == NULL) {
	return;
    }

    DmDatabaseInfo(p->spec.card_no, p->spec.db_id, db_name,
	    &attr, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
    StrCopy(arch_name, db_name);
    arch_name[dmDBNameLength - PZ_DB_EXT_LENGTH - 1] = '\0';
    StrCat(arch_name, PZ_DB_EXT);
    copy_protect = (attr & dmHdrAttrCopyPrevention) ? true : false;
    button = do_compress_dialog(
	    arch_name, &new, &copy_protect, &existing, &time);
    if (button != COMPRESS_OK_ID) {
	return;
    }

    db_id = DmFindDatabase(DEFAULT_CARD_NO, arch_name);
    if (db_id != NULL) {
	button = FrmCustomAlert(REPLACE_ALERT_ID, "archive", arch_name, " ");
	if (button != REPLACE_OK_ID) {
	    return;
	}
	DmDeleteDatabase(DEFAULT_CARD_NO, db_id);
    }

    form = start_progress(&saved_form, "Compressing...", db_name);
    start_ticks = TimGetTicks();
    if (new) {
	err = do_compress_new(DEFAULT_CARD_NO, arch_name, copy_protect,
		p->spec.card_no, p->spec.db_id, &total_size, &out_size);
    } else {
	err = do_compress_add(existing.card_no, existing.db_id, copy_protect,
		p->spec.card_no, p->spec.db_id, &total_size, &out_size);
    }
    end_ticks = TimGetTicks();
    end_progress(form, saved_form);

    if (err != noErr) {
	if (err != CANCELED) {
	    FrmCustomAlert(ERROR_ALERT_ID, "do_compress_db() failed", " ", " ");
	}
    } else if (time) {
	StrIToA(buf1, total_size);
	StrIToA(buf2, out_size);
	StrIToA(buf3, end_ticks - start_ticks);
	FrmCustomAlert(OK_ALERT_ID, buf1, buf2, buf3);
    }

    FrmUpdateForm(prefs.last_form_id, UPDATE_DB_LIST);
}
#endif


/*
 * getting information and opening
 */

static void do_info(void)
{
    struct db_rec *p;

    p = get_selected_db();
    if (p == NULL) {
	return;
    }

    if (get_db_info(p->spec.card_no, p->spec.db_id)) {
	FrmPopupForm(INFO_FORM_ID);
    }
}

static void do_open(void)
{
    struct db_rec *p;

    p = get_selected_db();
    if (p == NULL || p->kind != kind_archive) {
	return;
    }

    if (open_arch_db(p->spec.card_no, p->spec.db_id)) {
	FrmGotoForm(ARCH_FORM_ID);
    }
}

static void do_double_tap(void)
{
    struct db_rec *p;

    p = get_selected_db();
    if (p == NULL) {
	return;
    }

    if (p->kind == kind_archive) {
	do_open();
    } else {
	do_info();
    }
}


/*
 * event handlers
 */

Boolean handle_db_form_event(EventPtr event)
{
    int index;
    struct db_rec *p;
    Boolean handled = false;

    CALLBACK_PROLOGUE

    switch (event->eType) {
    case frmOpenEvent:
#ifndef UNPZ
	SetPopup(DB_SORT_POPUP_ID, DB_SORT_LIST_ID, prefs.sort_by);
	SetPopup(DB_INFO0_POPUP_ID, DB_INFO0_LIST_ID, prefs.info0);
	SetPopup(DB_INFO1_POPUP_ID, DB_INFO1_LIST_ID, prefs.info1);
	SetPopup(DB_INFO2_POPUP_ID, DB_INFO2_LIST_ID, prefs.info2);
	SetPopup(DB_WHERE_POPUP_ID, DB_WHERE_LIST_ID, prefs.where);
	SetPopup(DB_KIND_POPUP_ID, DB_KIND_LIST_ID, prefs.kind);
#endif
	FrmDrawForm(FrmGetActiveForm());
	draw_free_memory(DB_FREE_ID);
	init_table(DB_TABLE_ID, DB_SCROLLBAR_ID, draw_db_item);
	FrmUpdateForm(prefs.last_form_id, UPDATE_DB_LIST);
	handled = true;
	break;

    case frmUpdateEvent:
	draw_free_memory(DB_FREE_ID);
	switch (event->data.frmUpdate.updateCode) {
	case UPDATE_DB_LIST:
	case UPDATE_DB_SORT:
	case UPDATE_DB_FILTER:
	    update_db_list(event->data.frmUpdate.updateCode);
	    handled = true;
	    break;
	}
	break;

    case frmCloseEvent:
	free_db_list();
	break;

#ifndef UNPZ
    case popSelectEvent:
	CtlSetLabel(event->data.popSelect.controlP, LstGetSelectionText(
		event->data.popSelect.listP, event->data.popSelect.selection));
	switch (event->data.popSelect.listID) {
	case DB_SORT_LIST_ID:
	    prefs.sort_by = event->data.popSelect.selection;
	    FrmUpdateForm(prefs.last_form_id, UPDATE_DB_SORT);
	    break;

	case DB_INFO0_LIST_ID:
	    prefs.info0 = event->data.popSelect.selection;
	    goto do_redraw;
	case DB_INFO1_LIST_ID:
	    prefs.info1 = event->data.popSelect.selection;
	    goto do_redraw;
	case DB_INFO2_LIST_ID:
	    prefs.info2 = event->data.popSelect.selection;
	    goto do_redraw;
	do_redraw:
	    redraw_table(false);
	    break;

	case DB_WHERE_LIST_ID:
	    prefs.where = event->data.popSelect.selection;
	    goto do_update;
	case DB_KIND_LIST_ID:
	    prefs.kind = event->data.popSelect.selection;
	    goto do_update;
	do_update:
	    FrmUpdateForm(prefs.last_form_id, UPDATE_DB_FILTER);
	    break;
	}
	handled = true;
	break;
#endif

    case ctlSelectEvent:
	switch (event->data.ctlSelect.controlID) {
	case DB_ABOUT_ID:
	    FrmAlert(ABOUT_ALERT_ID);
	    handled = true;
	    break;

	case DB_INFO_ID:
	    do_info();
	    handled = true;
	    break;

#ifndef UNPZ
	case DB_COMPRESS_ID:
	    do_compress();
	    draw_free_memory(DB_FREE_ID);
	    handled = true;
	    break;
#endif

	case DB_OPEN_ID:
	    do_open();
	    handled = true;
	    break;
	}
	break;

    case keyDownEvent:
	if ((event->data.keyDown.modifiers & commandKeyMask) == 0
	 && (index = find_db(event->data.keyDown.chr)) >= 0) {
	    scroll_table(index - get_table_top(), true);
	    prefs.last_db_top = get_table_top();
	    handled = true;
	    break;
	}

	/* fall through */

    case sclRepeatEvent:
	handled = handle_table_event(event, NULL);
	prefs.last_db_top = get_table_top();
	break;

    case tblExitEvent:
	handled = handle_table_event(event, NULL);
	prefs.last_selection.db_id = NULL;
	break;

    case tblSelectEvent:
	handled = handle_table_event(event, do_double_tap);
	p = get_selected_db();
	if (p != NULL) {
	    prefs.last_selection = p->spec;
	}
	break;

    default:
	break;
    }

    CALLBACK_EPILOGUE

    return handled;
}
